# !/usr/bin/env python
"""Utility script with functions to be used with the results of GridSearchCV.

**plot_grid_search** plots as many graphs as parameters are in the grid search results.

**table_grid_search** shows tables with the grid search results.

Inspired in [Displaying the results of a Grid Search](https://www.kaggle.com/grfiv4/displaying-the-results-of-a-grid-search) notebook,
of [George Fisher](https://www.kaggle.com/grfiv4)
"""

import pandas as pd
from plotly.subplots import make_subplots
import plotly.graph_objs as go
import pprint
from scipy import stats
from IPython.display import display

__author__ = "Juanma Hernández"
__copyright__ = "Copyright 2019"
__credits__ = ["Juanma Hernández", "George Fisher"]
__license__ = "GPL"
__maintainer__ = "Juanma Hernández"
__email__ = "https://twitter.com/juanmah"
__status__ = "Utility script"


def plot_grid_search(clf):
    """Plot as many graphs as parameters are in the grid search results.

    Each graph has the values of each parameter in the X axis and the Score in the Y axis.

    Parameters
    ----------
    clf: estimator object result of a GridSearchCV
        This object contains all the information of the cross validated results for all the parameters combinations.
    """
    # Convert the cross validated results in a DataFrame ordered by `rank_test_score` and `mean_fit_time`.
    # As it is frequent to have more than one combination with the same max score,
    # the one with the least mean fit time SHALL appear first.
    cv_results = pd.DataFrame(clf.cv_results_).sort_values(by=['rank_test_score', 'mean_fit_time'])

    # Get parameters
    parameters=cv_results['params'][0].keys()

    # Calculate the number of rows and columns necessary
    rows = -(-len(parameters) // 2)
    columns = min(len(parameters), 2)
    # Create the subplot
    fig = make_subplots(rows=rows, cols=columns)
    # Initialize row and column indexes
    row = 1
    column = 1

    # For each of the parameters
    for parameter in parameters:

        # As all the graphs have the same traces, and by default all traces are shown in the legend,
        # the description appears multiple times. Then, only show legend of the first graph.
        if row == 1 and column == 1:
            show_legend = True
        else:
            show_legend = False

        # Mean test score
        mean_test_score = cv_results[cv_results['rank_test_score'] != 1]
        fig.add_trace(go.Scatter(
            name='Mean test score',
            x=mean_test_score['param_' + parameter],
            y=mean_test_score['mean_test_score'],
            mode='markers',
            marker=dict(size=mean_test_score['mean_fit_time'],
                        color='SteelBlue',
                        sizeref=2. * cv_results['mean_fit_time'].max() / (40. ** 2),
                        sizemin=4,
                        sizemode='area'),
            text=mean_test_score['params'].apply(
                lambda x: pprint.pformat(x, width=-1).replace('{', '').replace('}', '').replace('\n', '<br />')),
            showlegend=show_legend),
            row=row,
            col=column)

        # Best estimators
        rank_1 = cv_results[cv_results['rank_test_score'] == 1]
        fig.add_trace(go.Scatter(
            name='Best estimators',
            x=rank_1['param_' + parameter],
            y=rank_1['mean_test_score'],
            mode='markers',
            marker=dict(size=rank_1['mean_fit_time'],
                        color='Crimson',
                        sizeref=2. * cv_results['mean_fit_time'].max() / (40. ** 2),
                        sizemin=4,
                        sizemode='area'),
            text=rank_1['params'].apply(str),
            showlegend=show_legend),
            row=row,
            col=column)

        fig.update_xaxes(title_text=parameter, row=row, col=column)
        fig.update_yaxes(title_text='Score', row=row, col=column)

        # Check the linearity of the series
        # Only for numeric series
        if pd.to_numeric(cv_results['param_' + parameter], errors='coerce').notnull().all():
            x_values = cv_results['param_' + parameter].sort_values().unique().tolist()
            r = stats.linregress(x_values, range(0, len(x_values))).rvalue
            # If not so linear, then represent the data as logarithmic
            if r < 0.86:
                fig.update_xaxes(type='log', row=row, col=column)

        # Increment the row and column indexes
        column += 1
        if column > columns:
            column = 1
            row += 1

            # Show first the best estimators
    fig.update_layout(legend=dict(traceorder='reversed'),
                      width=columns * 360 + 100,
                      height=rows * 360,
                      title='Best score: {:.6f} with {}'.format(cv_results['mean_test_score'].iloc[0],
                                                                str(cv_results['params'].iloc[0]).replace('{',
                                                                                                          '').replace(
                                                                    '}', '')),
                      hovermode='closest',
                      template='none')
    fig.show()


def table_grid_search(clf, all_columns=False, all_ranks=False, save=True):
    """Show tables with the grid search results.

    Parameters
    ----------
    clf: estimator object result of a GridSearchCV
        This object contains all the information of the cross validated results for all the parameters combinations.

    all_columns: boolean, default: False
        If true all columns are returned. If false, the following columns are dropped:

        - params. As each parameter has a column with the value.
        - std_*. Standard deviations.
        - split*. Split scores.

    all_ranks: boolean, default: False
        If true all ranks are returned. If false, only the rows with rank equal to 1 are returned.

    save: boolean, default: True
        If true, results are saved to a CSV file.
    """
    # Convert the cross validated results in a DataFrame ordered by `rank_test_score` and `mean_fit_time`.
    # As it is frequent to have more than one combination with the same max score,
    # the one with the least mean fit time SHALL appear first.
    cv_results = pd.DataFrame(clf.cv_results_).sort_values(by=['rank_test_score', 'mean_fit_time'])

    # Reorder
    columns = cv_results.columns.tolist()
    # rank_test_score first, mean_test_score second and std_test_score third
    columns = columns[-1:] + columns[-3:-1] + columns[:-3]
    cv_results = cv_results[columns]

    if save:
        cv_results.to_csv('--'.join(cv_results['params'][0].keys()) + '.csv', index=True, index_label='Id')

    # Unless all_columns are True, drop not wanted columns: params, std_* split*
    if not all_columns:
        cv_results.drop('params', axis='columns', inplace=True)
        cv_results.drop(list(cv_results.filter(regex='^std_.*')), axis='columns', inplace=True)
        cv_results.drop(list(cv_results.filter(regex='^split.*')), axis='columns', inplace=True)

    # Unless all_ranks are True, filter out those rows which have rank equal to one
    if not all_ranks:
        cv_results = cv_results[cv_results['rank_test_score'] == 1]
        cv_results.drop('rank_test_score', axis = 'columns', inplace = True)        
        cv_results = cv_results.style.hide_index()

    display(cv_results)



import io
import mimetypes
import os
import shutil
import sys
import tarfile
import warnings
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import sklearn


def variable_memory():
    print('''
    def sizeof_fmt(num, suffix="B"):
        """by Fred Cirera,  https://stackoverflow.com/a/1094933/1870254, modified"""
        for unit in ["", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"]:
            if abs(num) < 1024.0:
                return "%3.1f %s%s" % (num, unit, suffix)
            num /= 1024.0
        return "%.1f %s%s" % (num, "Yi", suffix)
        
    
    for name, size in sorted(
        ((name, sys.getsizeof(value)) for name, value in locals().items()),
        key=lambda x: -x[1],
    )[:10]:
        print("{:>30}: {:>8}".format(name, sizeof_fmt(size)))
        '''
    )


def get_feature_names(column_transformer):
    """Get feature names from all transformers.
    Returns
    -------
    feature_names : list of strings
        Names of the features produced by transform.
    """
    # Remove the internal helper function
    # check_is_fitted(column_transformer)

    # Turn loopkup into function for better handling with pipeline later
    def get_names(trans):
        # >> Original get_feature_names() method
        if trans == "drop" or (hasattr(column, "__len__") and not len(column)):
            return []
        if trans == "passthrough":
            if hasattr(column_transformer, "_df_columns"):
                return (
                    column
                    if (
                        (not isinstance(column, slice))
                        and all(isinstance(col, str) for col in column)
                    )
                    else column_transformer._df_columns[column]
                )
            indices = np.arange(column_transformer._n_features)
            return ["x%d" % i for i in indices[column]]
        if not hasattr(trans, "get_feature_names"):
            # >>> Change: Return input column names if no method avaiable
            # Turn error into a warning
            warnings.warn(
                "Transformer %s (type %s) does not "
                "provide get_feature_names. "
                "Will return input column names if available"
                % (str(name), type(trans).__name__)
            )
            # For transformers without a get_features_names method, use the input
            # names to the column transformer
            return [] if column is None else [name + "__" + f for f in column]
        return [name + "__" + f for f in trans.get_feature_names()]

    ### Start of processing
    feature_names = []

    # Allow transformers to be pipelines. Pipeline steps are named differently, so preprocessing is needed
    if type(column_transformer) == sklearn.pipeline.Pipeline:
        l_transformers = [
            (name, trans, None, None)
            for step, name, trans in column_transformer._iter()
        ]
    else:
        # For column transformers, follow the original method
        l_transformers = list(column_transformer._iter(fitted=True))

    for name, trans, column, _ in l_transformers:
        if type(trans) == sklearn.pipeline.Pipeline:
            # Recursive call on pipeline
            _names = get_feature_names(trans)
            # if pipeline has no transformer that returns names
            if len(_names) == 0:
                _names = [name + "__" + f for f in column]
            feature_names.extend(_names)
        else:
            feature_names.extend(get_names(trans))

    return feature_names


def dir_walk(directory, show_hidden_dir=False, exclude=None, include=None):
    """
    Walks through a directory and its subdirectories recursively and counts the number of files of each MIME type.

    Args:
        directory (str): The path to the directory.
        show_hidden_dir (bool): Whether to show hidden directories or not.
        exclude (list): A list of folder names to exclude.
        include (list): A list of folder names to include.

    Returns:
        pd.DataFrame: A dataframe containing the file counts.
    """
    if exclude is None:
        exclude = []
    if include is None:
        include = []
    final = []

    for root, dirs, files in os.walk(directory):
        dir_name = os.path.basename(root)

        # Exclude hidden directories
        if not show_hidden_dir:
            dirs[:] = [d for d in dirs if not d.startswith('.')]

        # Exclude directories in the exclude list
        if dir_name in exclude:
            dirs.clear()
            continue

        # Include only directories in the include list
        if include and dir_name not in include:
            # dirs.clear()
            continue

        file_counts = {
            'directory': root,
            'base': dir_name,
            'folders': len(dirs),
            'video': 0,
            'music': 0,
            'photos': 0,
            'application/zip': 0,
            'documents': 0,
            'others':0,
            'total files':len(files)
        }

        for file in files:
            file_path = os.path.join(root, file)
            mime_type, _ = mimetypes.guess_type(file_path)
            if mime_type is None:
                file_counts['others'] += 1

            elif mime_type.startswith('video/'):
                file_counts['video'] += 1
            elif mime_type.startswith('audio/'):
                file_counts['music'] += 1
            elif mime_type.startswith('image/'):
                file_counts['photos'] += 1
            elif mime_type == 'application/zip':
                file_counts['application/zip'] += 1
            elif mime_type.startswith('text/') or mime_type.endswith('msword') or mime_type.endswith('pdf'):
                file_counts['documents'] += 1
        final.append(file_counts)

    return pd.DataFrame(final)


def printt(*args, sep=' ', end='\n',terminal_width:int=170, file=sys.stdout, flush=False):
    
    if terminal_width is None:
        terminal_width = shutil.get_terminal_size().columns

    # Calculate the available width for each argument
    num_args = len(args)
    arg_width = terminal_width // num_args

    # Determine if there are any plots in the arguments
    has_plots = any(isinstance(arg, plt.Figure) for arg in args)

    # Determine the number of lines needed for each argument
    max_lines = max(
        max(str(arg).count('\n') + 1 for arg in args),
        1 if has_plots else 0
    )

    # Loop through each line
    for line_num in range(max_lines):
        # Loop through each argument
        for i, arg in enumerate(args):
            # Convert argument to string
            arg_str = str(arg)

            # Check if the argument is a plot
            if isinstance(arg, plt.Figure):
                plt.figure(arg.number)
                plt.imshow(arg)
                plt.axis('off')
                plt.show()
                continue

            # Split the argument string into lines
            lines = arg_str.splitlines()

            # Get the current line or an empty string if there are no more lines
            line = lines[line_num] if line_num < len(lines) else ""

            # Calculate the width for this argument
            arg_width_adjusted = arg_width - len(sep)
            if i == num_args - 1:
                arg_width_adjusted -= len(end)

            # Truncate or pad the line to fit the width
            line_str = line[:arg_width_adjusted].ljust(arg_width_adjusted)

            # Print the line
            file.write(line_str)

            # Print the separator between arguments
            if i < num_args - 1:
                file.write(sep)

        # Print the end character/string
        file.write(end)

    # Flush the output if specified
    if flush:
        file.flush()


import json

import toml
import yaml
from yaml import SafeLoader
from yaml.constructor import ConstructorError


class DotDict(dict):
    def __init__(self, dictionary=None):
        if dictionary is not None:
            for key, value in dictionary.items():
                if isinstance(value, dict):
                    value = DotDict(value)
                self[key] = value


    def __getattr__(self, key):
        if key in self:
            return self[key]
        else:
            raise AttributeError(f"'DotDict' object has no attribute '{key}'")

    def __setattr__(self, key, value):
        self[key] = value


    def save_json(self, file_path):
        with open(file_path, 'w') as file:
            json.dump(self, file, indent=4)

    def save_toml(self, file_path):
        with open(file_path, 'w') as file:
            toml.dump(self, file)

    def save_yaml(self, file_path):
        with open(file_path, 'w') as file:
            yaml.dump(self, file)

    @classmethod
    def from_json(cls, file_path):
        with open(file_path, 'r') as file:
            data = json.load(file)
        return cls(data)

    @classmethod
    def from_toml(cls, file_path):
        with open(file_path, 'r') as file:
            data = toml.load(file)
        return cls(data)



if __name__ == '__main__':
    print('main')


def main():
    ...


if __name__ == "__main__":

    main()



import pandas as pd
import matplotlib.pyplot as plt
import numpy as np
import random

class WindowData:
    '''
    dataclass returns windows and horizons with given parameter

    -------------------
    Parameters:
        data:nd array       :   numpy array or pandas dataframe
        window_size:int     :   size of the window
        horizon_size:int    :   size of the horizon

        stride:int          :   how much window slides on data
        shift:int           :   gap between window and horizon(if needed) 

    '''

    def __init__(self, data: pd.DataFrame , window_size:int=7, horizon_size:int=1, label_columns_indices:list=None,label_columns_names:list = None, stride:int = 1, shift:int=0):
        self.data = data
        self.window_size = window_size
        self.horizon_size = horizon_size
        self.label_columns_indices = label_columns_indices
        self.label_columns_names = label_columns_names
        self.stride = stride
        self.shift = shift




        ## getting label indices
        if type(self.data) == pd.core.frame.DataFrame:

            self.column_indices = {name: i for i, name in
                                enumerate(self.data.columns)}

            if self.label_columns_names is not None:
                self.label_columns_indices = [self.column_indices[name]  for name in
                                        self.label_columns_names]
            
            # converting dataframe to ndarray
            self.data = self.data.to_numpy()


        # create a array of window size
        self.window_size_array = np.arange(self.window_size + self.shift + self.horizon_size)

        # create array with a forward shift
        self.all_array_index = np.arange(len(self.data)- (self.window_size + self.shift + self.horizon_size -1), step = self.stride).reshape(-1,1) + self.window_size_array

        # Spliting index in window and horizon
        self.all_windows_index = self.all_array_index[:, :self.window_size]
        self.all_horizons_index = self.all_array_index[:, self.window_size+self.shift:]


        # actuall values of horizon and windows
        self.all_windows = self.data[self.all_windows_index]
        self.all_horizons_full = self.data[self.all_horizons_index] # contains all the labels

        # Work out the label column indices.
        if self.label_columns_indices is not None:
            self.all_horizons = self.all_horizons_full[...,self.label_columns_indices]
        
        else:
            self.all_horizons = self.all_horizons_full


    def make_split(self):
        return self.all_windows, self.all_horizons

    def train_test_val(self, test_size:float=.1, val_size:float=None, shuffle:bool=True):
        '''
        returns  tuple of train and test or train, test, val

        test_size: float = 0.1
        val_size : float = None
        shuffle  : bool  = True

        if val_size is not None:
            train, test = train_test_val()
        else:
            train, test, val = train_test_val() 
        
        '''
        x, y = self.make_split()   

        test_len = int(len(self.all_windows)* test_size)
        
        # shuffle
        np.random.shuffle(test_len) if shuffle else test_len

        xtrainrest, xtest, ytrainrest, ytest = self.all_windows[:-test_len], self.all_windows[-test_len:], self.all_horizons[:-test_len], self.all_horizons[-test_len:]
        if val_size is None:
            return (xtrainrest, ytrainrest), (xtest, ytest)
        else:
            val_len = int(len(xtrainrest)* val_size)
            xtrain, xval, ytrain, yval = xtrainrest[:-val_len], xtrainrest[-val_len:], ytrainrest[:-val_len], ytrainrest[-val_len:]
            return (xtrain, ytrain), (xtest, ytest), ( xval, yval)

    def plot(self, plot_cols:list=None, model = None, figsize = (15,5),**kwargs):

        # plots data
        plt.figure(figsize = figsize)
        
        for i, name in enumerate(plot_cols):
            plt.subplot(len(plot_cols),1, i+1)
            plt.grid(True)
            plt.plot(self.all_horizons_full[...,self.column_indices[name]],  **kwargs)
            plt.title(name)
            # if model is not None:
            #     predictions = model.pred
            # plt.plot(self.all_horizons_full[...,self.columns_indices[name]])
        
        plt.tight_layout()
        plt.show()




import json
import os
import random
import re
import shutil
from pathlib import Path


# make data splitstr
def make_data_split(
    main_path,
    test=True,
    test_split_ratio: float = 0.2,
    val=True,
    val_split_ratio: float = 0.1,
    shuffle=True,
):
    """
    Make data split Train/test/val
    Note: keep a copy of the ORIGINAL DATASET

    Args:
        main_path = Dataset directory that has class subdirectory
        test_split_ratio = float value between 0 to 1
        val = bool to split for validation data
        val_split_ratio = float value for split
        shuffle = shuffle

    Returns:
        Splited DATASET (train/test/validation)

    """
    main = Path(main_path)
    class_names = sorted([i.name for i in main.iterdir()])
    print(class_names)
    total_files = len(os.listdir(main / class_names[0]))
    try:
        # create train/test/validation
        os.makedirs(main / "train")

        # creating test set
        if test:
            os.makedirs(main / "test")
            test_image_num = 0
            test_image_num = int(total_files * test_split_ratio)
            print("test images:", test_image_num)
            for class_name in class_names:
                class_path = main / class_name
                if shuffle:
                    sample_images = random.sample(
                        os.listdir(class_path), test_image_num
                    )

                else:
                    sample_images = sorted(os.listdir(class_path))[:test_image_num]

                sample_images = [class_path / i for i in sample_images]
                test_class = str(main / "test" / class_name)
                os.makedirs(test_class)
                for file in sample_images:
                    shutil.move(str(file), test_class)

        # creating validation dataset
        val_image_num = 0
        if val:
            os.makedirs(main / "validation")
            val_image_num = int(total_files * val_split_ratio)
            print("val images:", val_image_num)
            for class_name in class_names:
                class_path = main / class_name
                if shuffle:
                    sample_images = random.sample(os.listdir(class_path), val_image_num)

                else:
                    sample_images = sorted(os.listdir(class_path))[:val_image_num]

                sample_images = [class_path / i for i in sample_images]
                test_class = str(main / "validation" / class_name)
                os.makedirs(test_class)
                for file in sample_images:
                    shutil.move(str(file), test_class)

        for class_name in class_names:
            shutil.move(str(main / class_name), str(main / "train" / class_name))

        print("train images:", total_files - test_image_num - val_image_num)
    except Exception as e:
        print("Exception Occured: ", e)

def download_kaggle_resource(resource, download_path="dataset", kaggle_json_path=None):
    """
    Downloads a Kaggle dataset or competition using the Kaggle API key stored in kaggle.json.

    Args:
        resource (str): URL of the Kaggle dataset/competition or API command.
        download_path (str): Path where the resource will be downloaded.
        kaggle_json_path (str, optional): Path to the kaggle.json file containing the API key.
    """

    # Find the path to kaggle.json if not provided
    default_path = Path(os.path.expanduser("~/.kaggle/kaggle.json"))
    default_path.parent.mkdir(exist_ok=True)
    if not default_path.exists():
        if kaggle_json_path is not None:
            shutil.copy(kaggle_json_path, default_path)
        else:
            raise ValueError(
                "kaggle.json not found in default location and kaggle_json_path not provided."
            )


    # Load the Kaggle API key from kaggle.json
    with open(default_path) as f:
        kaggle_credentials = json.load(f)

    # Configure the Kaggle API
    from kaggle.api.kaggle_api_extended import KaggleApi

    api = KaggleApi()
    api.authenticate()

    if resource.startswith("kaggle "):
        # Run the provided API command directly
        os.system(resource)
        return

    # Handle URL-based resource
    url_parts = resource.split("/")

    if "datasets" in url_parts:
        resource_type = "datasets"
        dataset_part_index = url_parts.index("datasets") + 1
    elif "competitions" in url_parts:
        resource_type = "competitions"
        competition_part_index = url_parts.index("competitions") + 1
    else:
        raise ValueError("Invalid Kaggle URL")

    # Extract resource name from the URL
    resource_name = url_parts[-1]

    # Define the download path for the resource
    resource_download_path = Path(download_path) / resource_name

    # Create the directory if it doesn't exist
    resource_download_path.mkdir(parents=True, exist_ok=True)

    # Download the resource
    if resource_type == "datasets":
        dataset_path = "/".join(url_parts[dataset_part_index:])
        api.dataset_download_files(dataset_path, path=resource_download_path)
    elif resource_type == "competitions":
        competition_path = "/".join(url_parts[competition_part_index:])
        api.competition_download_files(competition_path, path=resource_download_path)

    print(f"Resource downloaded to: {resource_download_path}")


if __name__ == "__main__":
    print('hello world')



import os 
from tqdm import tqdm
from PIL import Image

def resize_func( file_path:str, dest_path:str = None ,px:int  = 512, keep_aspect_ratio = True, quality:int = 100, format = None):
    '''
    Resizes photos in a directory with given paramters
    
    -----------------------------
    
    Paramters-----------------
        file_path:str =  folder which contains files
        
        dest_path:str =  folder where to keep the resized folder (if not provided
                        resized files will be kept in file_path)
                        
        px:int = max pixel range to resize upto

        keep_aspect_ratio =  keep aspect ratio ?

        format:str = default: None ['PNG','JPEG'] any one
        
        quality:int = quality range between 1 to 100  
    '''
    
    if dest_path == None:
        dest_path = file_path
    
    try:
        for item in tqdm(os.listdir(file_path)):
            img_path = os.path.join(file_path, item)        
            img = Image.open(img_path)
            #Dimensions
            w, h = img. size
#             print('height:', h, 'width: ', w)
            #ratio
            ratio = h/w
#             print(ratio)

            filename, extension = os.path.splitext(item)
#             print(filename, '   :  ', extension)


            if keep_aspect_ratio:        
                if ratio < 1:
                    h = px * ratio
                    w = px 


                else:
                    w = px/ratio
                    h = px 
            else:
                h = px
                w = px

            if not os.path.exists(dest_path):
                os.makedirs(dest_path)

#             print('new height:', h, 'new width:',w)

            img_resized = img.resize((int(w),int(h)), Image.ANTIALIAS)
            img_resized.save(dest_path + f'/{filename}_resized{extension}', format = format, quality=100)

    except Exception as e:
        raise ('Exception Occured: ',e)

def main(): 
    ...


if __name__ == '__main__':
    main()



import json
import re
from collections import Counter
from functools import cache
from typing import Callable, Generator, List

__all__ = ['contracations_dict', 'text_cleaning', 'IntegerVectorizer', 'Vocabulary', 'fix_contractions']
# with open('contractions.json') as f:
#    contractions_dict = json.load(f)


contractions_dict = { 
"i ain't": "i am not ", 
"you ain't":"you are not" ,
"he ain't":"he is not", 
"she ain't":"she is not",
"it ain't":"it is not",
"they ain't":"they are not",
"aren't": "are/am not",
"can't": "cannot",
"can't've": "cannot have",
"'cause": "because",
"could've": "could have",
"couldn't": "could not",
"couldn't've": "could not have",
"didn't": "did not",
"didn`t": "did not",
"doesn't": "does not",
"don't": "do not",
"hadn't": "had not",
"hadn't've": "had not have",
"hasn't": "has not",
"haven't": "have not",
"he'd": "he had / he would",
"he'd've": "he would have",
"he'll": "he shall / he will",
"he'll've": "he shall have / he will have",
"he's": "he has / he is",
"how'd": "how did",
"how'd'y": "how do you",
"how'll": "how will",
"how's": "how has / how is / how does",
"I'd": "I had / I would",
"I'd've": "I would have",
"I'll": "I shall / I will",
"I'll've": "I shall have / I will have",
"I'm": "I am",
"I've": "I have",
"isn't": "is not",
"it'd": "it had / it would",
"it'd've": "it would have",
"it'll": "it shall / it will",
"it'll've": "it shall have / it will have",
"it's": "it has / it is",
"let's": "let us",
"ma'am": "madam",
"mayn't": "may not",
"might've": "might have",
"mightn't": "might not",
"mightn't've": "might not have",
"must've": "must have",
"mustn't": "must not",
"mustn't've": "must not have",
"needn't": "need not",
"needn't've": "need not have",
"o'clock": "of the clock",
"oughtn't": "ought not",
"oughtn't've": "ought not have",
"shan't": "shall not",
"sha'n't": "shall not",
"shan't've": "shall not have",
"she'd": "she had / she would",
"she'd've": "she would have",
"she'll": "she shall / she will",
"she'll've": "she shall have / she will have",
"she's": "she has / she is",
"should've": "should have",
"shouldn't": "should not",
"shouldn't've": "should not have",
"so've": "so have",
"so's": "so as / so is",
"that'd": "that would / that had",
"that'd've": "that would have",
"that's": "that has / that is",
"there'd": "there had / there would",
"there'd've": "there would have",
"there's": "there has / there is",
"they'd": "they had / they would",
"they'd've": "they would have",
"they'll": "they shall / they will",
"they'll've": "they shall have / they will have",
"they're": "they are",
"they've": "they have",
"to've": "to have",
"wasn't": "was not",
"we'd": "we had / we would",
"we'd've": "we would have",
"we'll": "we will",
"we'll've": "we will have",
"we're": "we are",
"we've": "we have",
"weren't": "were not",
"what'll": "what shall / what will",
"what'll've": "what shall have / what will have",
"what're": "what are",
"what's": "what has / what is",
"what've": "what have",
"when's": "when has / when is",
"when've": "when have",
"where'd": "where did",
"where's": "where has / where is",
"where've": "where have",
"who'll": "who shall / who will",
"who'll've": "who shall have / who will have",
"who's": "who has / who is",
"who've": "who have",
"why's": "why has / why is",
"why've": "why have",
"will've": "will have",
"won't": "will not",
"won't've": "will not have",
"would've": "would have",
"wouldn't": "would not",
"wouldn't've": "would not have",
"y'all": "you all",
"y'all'd": "you all would",
"y'all'd've": "you all would have",
"y'all're": "you all are",
"y'all've": "you all have",
"you'd": "you had / you would",
"you'd've": "you would have",
"you'll": "you shall / you will",
"you'll've": "you shall have / you will have",
"you're": "you are",
"you've": "you have"
}

# very necessary function you can't tell differece between "’" and "'", I don't even have that symbol in my keyboard
def text_cleaning_apos(text):
    text = str(text)
    text = text.lower()
    text = re.sub("’", "'", text) # removing punctuation
    return text


# FUNCTIONS TO EXPAND CONTRACTIONS
def fix_contractions(x):
    x = str(x).lower()
    xsplited = x.split(' ')
    exp_sentence = []
    for s in x.split():
        if s in contractions_dict.keys():

            s = contractions_dict.get(s)
        exp_sentence.append(s)

    return ' '.join(exp_sentence)

# Compile regex patterns for better performance
PUNCTUATION_REGEX = re.compile(r'[^a-zA-Z]')
SPECIAL_CHARACTERS_REGEX = re.compile(r'[#,@,&]')
DIGIT_REGEX = re.compile(r'\d+')
APOSTROPHE_S_REGEX = re.compile(r"'s")
WWW_REGEX = re.compile(r'w{3}')
URL_REGEX = re.compile(r'http\S+')
MULTIPLE_SPACES_REGEX = re.compile(r'\s+')
SINGLE_CHAR_REGEX = re.compile(r'\s+[a-zA-Z]\s+')

def text_cleaning(text):
    text = str(text)
    text = text.lower()
    text = PUNCTUATION_REGEX.sub(' ', text)  # Remove punctuation
    text = SPECIAL_CHARACTERS_REGEX.sub('', text)
    text = DIGIT_REGEX.sub('', text)
    text = APOSTROPHE_S_REGEX.sub('', text)
    text = WWW_REGEX.sub('', text)
    text = URL_REGEX.sub('', text)
    text = MULTIPLE_SPACES_REGEX.sub(' ', text)
    text = SINGLE_CHAR_REGEX.sub(' ', text)
    
    return text.strip()


class Vocabulary:
    
    """
    Class for storing a vocabulary of words and their corresponding indices.

    Args:
        special_tokens: A list of special tokens to add to the vocabulary .
        Note: special_tokens will not be processed by preprocessing function.

    Attributes:
        word_to_idx: A dictionary mapping words to their indices.
        idx_to_word: A dictionary mapping indices to their words.
        counter: A counter of the number of times each word appears in the data.
        UNK_TOKEN: The token for unknown words.
        UNK: The index of the UNK_TOKEN.
        PAD_TOKEN: The token for padding.
        PAD: The index of the PAD_TOKEN.
        vocab_size: The size of the vocabulary.

    Methods:
        build_vocab: Build the vocabulary from a dataset of text.
        add_word_to_vocab: Add a word to the vocabulary.
        __len__: Get the length of the vocabulary.
    """
    
    def __init__(self, special_tokens: list[str]):
        self.word_to_idx = {}
        self.idx_to_word = {}
        self.counter = Counter()
        
        self.UNK_TOKEN = '<UNK>'
        self.UNK = 1
        self.PAD_TOKEN = '<PAD>'
        self.PAD = 0
        
        self.word_to_idx[self.UNK_TOKEN] = self.UNK
        self.idx_to_word[self.UNK] = self.UNK_TOKEN

        self.word_to_idx[self.PAD_TOKEN] = self.PAD
        self.idx_to_word[self.PAD] = self.PAD_TOKEN

        self.vocab_size = 2
        for idx, token in enumerate(special_tokens, start=2):
            self.word_to_idx[token] = idx
            self.idx_to_word[idx] = token
            self.vocab_size += 1

    def build_vocab(self, tokenized_data, max_tokens, min_freq):
        # sourcery skip: dict-assign-update-to-union
        self.counter = Counter()
        for words in tokenized_data:
            self.counter.update(words)

        if max_tokens is not None:
            sorted_tokens = [word for word, _ in self.counter.most_common()]
            for word in sorted_tokens:
                if word not in self.word_to_idx:
                    self.add_word_to_vocab(word)

                if self.vocab_size == max_tokens:
                    break
        else:
            for word, freq in self.counter.items():
                if freq >= min_freq and word not in self.word_to_idx:
                    self.add_word_to_vocab(word)

    def add_word_to_vocab(self, word):
        self.word_to_idx[word] = self.vocab_size
        self.idx_to_word[self.vocab_size] = word
        self.vocab_size += 1

    def __len__(self):
        return self.vocab_size
    

class IntegerVectorizer:
    """
    Class for converting text data to integer vectors.

    Args:
        - tokenizer: A function that takes a string and returns a list of tokens. (default=None)
        - preprocessing_func: A function that takes a token and returns a processed token. (default=None)
        - max_tokens: The maximum number of tokens to keep in the vocabulary. (default=None)
        - min_freq: The minimum frequency of a token to keep in the vocabulary. (default=1)
        - special_tokens: A list of special tokens to add to the vocabulary. (default=None)
        - max_seq_length: The maximum sequence length for each input. (default=None)
        - pad_to_max: Whether to pad sequences to the maximum length. (default=False)

    Attributes:
        vocab: The vocabulary used to convert text to integers.
        tokenized_data: The tokenized data used to build the vocabulary.

    Methods:
        - adapt: Adapt the vectorizer to a dataset of text.
        - __call__: Convert text data to integer vectors.
        - preprocess_sentence: Preprocess a sentence for tokenization.
        - tokenize_data_generator: Tokenize a dataset of text and yield the tokens.
        - transform: Convert a dataset of text to integer vectors.
        - adjust_sequence_length: Adjust the length of a sequence to the maximum length.
        - reverse_transform: Convert integer vectors to text.
        - transform_generator: Convert a dataset of text to integer vectors in a generator.
        - reverse_transform_generator: Convert integer vectors to text in a generator.
    """
    
    def __init__(self, 
                 tokenizer: Callable[[str], list[str]] = None,
                 preprocessing_func: Callable[[str], str] = None,
                 max_tokens=None,
                 min_freq=1,
                 special_tokens: list[str] = None,
                 max_seq_length=None,
                 pad_to_max=False):
        
        self.min_freq = min_freq
        self.max_tokens = max_tokens
        self.max_seq_length = max_seq_length
        self.tokenizer = tokenizer
        self.preprocessing_func = preprocessing_func
        self.reserved_tokens = ['<UNK>', '<PAD>']
        self.special_tokens = [token for token in special_tokens if token not in self.reserved_tokens] if special_tokens else []
        self.pad_to_max = pad_to_max  # Store the argument

        self.vocab = Vocabulary(self.special_tokens)
        self.tokenized_data = []

    def adapt(self, data):
        self.tokenized_data = self.tokenize_data_generator(data)
        self.vocab.build_vocab(self.tokenized_data, self.max_tokens, self.min_freq)
        print('Vocab size:', len(self.vocab))

    def __call__(self, data, reverse=False, return_generator = True):
        if reverse:
            return self.reverse_transform_generator(data) if return_generator else self.reverse_transform(data)
        else:
            return self.transform_generator(data) if return_generator else self.transform(data)

    def preprocess_sentence(self, sentence):
        if self.preprocessing_func:
            words = sentence.split()
            preprocessed_words = [self.preprocessing_func(word) if word not in self.special_tokens else word for word in words]
            return " ".join(preprocessed_words)
        return sentence

    def tokenize_data_generator(self, data):
        for sentence in data:
            sentence = self.preprocess_sentence(sentence)
            yield self.tokenizer(sentence) if self.tokenizer else str(sentence).split()

    def transform(self, data:List[str]):
        
        if not isinstance(data, list):
            raise TypeError("Input data must be a list")
        
        self.tokenized_data = self.tokenize_data_generator(data)
        vectorized_data = []
        for sentence in self.tokenized_data:
            vectorized_sentence = [self.vocab.word_to_idx.get(word, self.vocab.UNK) for word in sentence]
            vectorized_sentence = self.adjust_sequence_length(vectorized_sentence)
            vectorized_data.append(vectorized_sentence)
        return vectorized_data

    def adjust_sequence_length(self, sequence: Generator[int, None, None]) -> list[int]:
        if self.max_seq_length is not None:
            
            if isinstance(sequence, Generator):                
                sequence = list(sequence)
                
            if len(sequence) < self.max_seq_length:
                if self.pad_to_max:
                    sequence += [self.vocab.PAD] * (self.max_seq_length - len(sequence))
            elif len(sequence) > self.max_seq_length:
                sequence = sequence[:self.max_seq_length]
            return sequence

    def reverse_transform(self, vectorized_data: list[list[int]]) -> list[str]:
        original_data = []
        for vector in vectorized_data:
            sentence = [self.vocab.idx_to_word[idx] for idx in vector if idx != self.vocab.PAD]
            original_data.append(" ".join(sentence).strip())
        return original_data


    def transform_generator(self, data: list[str]) -> Generator[list[int], None, None]:
        
        
        if not isinstance(data, list):
            raise TypeError("Input data must be a list")
        
        self.tokenized_data = self.tokenize_data_generator(data)
        for sentence in self.tokenized_data:
            vectorized_sentence = (self.vocab.word_to_idx.get(word, self.vocab.UNK) for word in sentence)
            vectorized_sentence = self.adjust_sequence_length(vectorized_sentence)
            yield list(vectorized_sentence)  # Convert the generator to a list for yielding

    def reverse_transform_generator(self, vectorized_data: list[list[int]]) -> Generator[str, None, None]:
        for vector in vectorized_data:
            sentence = (self.vocab.idx_to_word[idx] for idx in vector if idx != self.vocab.PAD)
            yield " ".join(sentence).strip()


def main():
	...
	
if __name__ == '__main__':
	main()



from sklearn.metrics import accuracy_score, precision_recall_fscore_support
from sklearn.metrics import f1_score, precision_score, recall_score,roc_auc_score, confusion_matrix
import seaborn as sns
    


def all_metrics(y_true, y_pred, heatmap = True):
    import os
    os.system('pip install scikit-learn seaborn')
  
    print('f1 score: ',f1_score(y_true, y_pred))
    print( 'precision score: ', precision_score(y_true, y_pred))
    print('recall score: ',recall_score(y_true,y_pred))
    auc = roc_auc_score(y_true, y_pred)
    print('AUC: %.3f' % auc)
    if heatmap:
        sns.heatmap(confusion_matrix(y_true, y_pred), annot=True,cmap='Blues', fmt='g')
    else:
        print(confusion_matrix(y_true, y_pred))


# Function to evaluate: accuracy, precision, recall, f1-score
def calculate_results(y_true, y_pred, model_name:str=None, discription:str = None):
    """
  Calculates model accuracy, precision, recall and f1 score of a binary classification model.
  Args:
      y_true: true labels in the form of a 1D array
      y_pred: predicted labels in the form of a 1D array
  Returns a dictionary of accuracy, precision, recall, f1-score.
  """
    # Calculate model accuracy
    model_accuracy = accuracy_score(y_true, y_pred) * 100
    # Calculate model precision, recall and f1 score using "weighted average
    model_precision, model_recall, model_f1, _ = precision_recall_fscore_support(y_true, y_pred, average="weighted")
    return {
        "model": model_name,
        "accuracy": model_accuracy,
        "precision": model_precision,
        "recall": model_recall,
        "f1": model_f1,
        "discription": discription,
    }



def main():
	pass


if __name__=="__main__":
	
	main()



import tensorflow as tf
from tensorflow import keras
import os
import datetime
import random


def create_model_checkpoint(model_name, save_dir, monitor:str = 'val_loss',verbose: int = 0, save_best_only: bool = True, save_weights_only: bool = False,
                            mode: str = 'auto', save_freq='epoch', options=None, initial_value_threshold=None, **kwargs):
	"""
	Creates a ModelCheckpoint callback instand to store model weights.
	Stores model weights with the filepath:
	"save_dir/model_name-{str(datetime.datetime.now())}"
	
	Args
	###########################################
	model_name: name of model directory (e.g. efficientnet_model_1)

	save_dir: target directory to store model weights
	
	monitor: metric to monitor
	
	verbose: verbosity mode
	
	save_best_only: if True, only save the model if `monitor` has improved.
	
	save_weights_only: if True, then only the model's weights will be saved
	
	mode: one of {auto, min, max}. In `min` mode, the model will be minimized
	    by the quantity monitored, in `max` mode the model will be maximized
	    (by the quantity monitored). In `auto` mode, the direction is
	    automatically inferred from the name of the monitored quantity.
	
	save_freq: one of {epoch, batch}. In `epoch` mode, the callback saves the
	"""
	model_name = f'{model_name}-{str(datetime.datetime.now())}'
	directory = os.path.join(save_dir, model_name)

	if not os.path.exists(directory):
		os.makedirs(directory)

	return tf.keras.callbacks.ModelCheckpoint(
												directory,
												monitor = monitor,
												verbose = verbose,
												save_best_only = save_best_only,
												save_weights_only = save_weights_only,
												mode = mode,
												save_freq = save_freq,
												options=options,
												initial_value_threshold = initial_value_threshold,
												**kwargs)





def create_tensorboard_callback(dir_name, experiment_name):
    """
	Creates a TensorBoard callback instand to store log files.
	Stores log files with the filepath:
	"dir_name/experiment_name/current_datetime/"
	
	Args
	###########################################
	dir_name: target directory to store TensorBoard log files
	experiment_name: name of experiment directory (e.g. efficientnet_model_1)
	
	Returns
	##########################################
	A tensorboard callback
	"""
    log_dir = (
        f"{dir_name}/{experiment_name}/"
        + datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
    )
    tensorboard_callback = tf.keras.callbacks.TensorBoard(
      log_dir=log_dir
    )
    print(f"Saving TensorBoard log files to: {log_dir}")
    return tensorboard_callback


class DisplayCallback(tf.keras.callbacks.Callback):
	'''
	Prints model prediction and true values for dataset specified
	
	Args
	###########################################
	print_on_epoch:int : after how many epochs you want to print the results
	x:np.array  : input dataset (np.array)
	y:np.array  : output dataset (np.array)
	
	Returns
	##########################################
	A Display callback
	
	'''
	def __init__(self,print_on_epoch:int, x:np.array, y:np.array):
		self.print_on_epoch = print_on_epoch
		self.x = x
		self.y = y
		
	def predict(self):
		ypred = tf.squeeze(model.predict(self.x))
		print('ypred: ',ypred[0].numpy().astype('int'),'\nytrue: ',self.y[0])

	def on_epoch_end(self, epoch, logs=None):
		if (epoch%self.print_on_epoch==0):
			return self.predict()



def main():
    ...

    
if __name__ == '__main__':
	main()




from setuptools import setup, find_packages

setup(
    name='funcyou',
    version='1.0.6',
    author='Tikendra sahu',
    author_email='tikendraksahu1029@gmail.com',
    description='useful function to save time',
    packages=find_packages(),
    install_requires=[
        'matplotlib',
        'scikit-learn',
        'pandas',
        'numpy',
        'kaggle'
    ],
)



import matplotlib as mpl
import tensorflow as tf



		
def create_bert(trainable=False):
    import tensorflow_hub as hub
    text_input = tf.keras.layers.Input(shape=(), dtype=tf.string)
    preprocessor = hub.KerasLayer(
        "https://tfhub.dev/tensorflow/bert_en_uncased_preprocess/3")
    encoder_inputs = preprocessor(text_input)
    encoder = hub.KerasLayer(
        "https://tfhub.dev/tensorflow/small_bert/bert_en_uncased_L-12_H-128_A-2/2",
        trainable=trainable)
    outputs = encoder(encoder_inputs)
    pooled_output = outputs["pooled_output"]      
    sequence_output = outputs["sequence_output"]  
    embedding_model = tf.keras.Model(text_input, pooled_output)
    return embedding_model
		
		
if __name__=='__main__':
	print('hello')



import os , random, shutil
from pathlib import Path
import matplotlib.pyplot as plt
import matplotlib.image as mpimg
import seaborn as sns
import pandas as pd
from PIL import Image
import numpy as np
import matplotlib.image as mpimg

				
from sklearn.metrics import confusion_matrix
import itertools




def plot_random_dataset(main_path, row:int = 2, col:int = 5, figsize:tuple= (15,6), recursive = True):

    '''
    Plots random images from the Path provided
    
    Args:
      main_path:str = path to the directory
      row:int       = number of rows needed
      col:int		  = number of columns needed
      figsize		 = size of the figure 

      Returns:
          A figure of row*cols of random images from dataset

    '''
    main_path = Path(main_path)
    all_images = []
    image_extension = ['jpg','jpeg','png','tiff','webp','svg','pjpeg']  

    if recursive:# walk through all dir and get images
        for dir, folders, files in os.walk(main_path):
            filles = [f'{dir}/{i}' for i in files if i.split('.')[-1] in image_extension]
            all_images += filles
    else:
        for i in main_path.iterdir():
          if i.suffix in image_extension:
            all_images.append(i)
    #sampling random images 
    sample_images = random.sample(all_images, row*col)

    # plotting figure
    fig = plt.figure(figsize = figsize)

    if len(all_images) <= 0:
      print('No Image Found', len(all_images))

    for i,image in enumerate(sample_images):
        title = str(image).split('/')[-2]
        img = mpimg.imread(image)
        plt.subplot(row, col, i+1)
        plt.title(title)
        plt.imshow(img)

    plt.show()
	
# function to plot an image and details
def plot_image_with_details(image_path):
    print(image_path)
    plt.subplot(1,2,1)
    image_ = Image.open(image_path)
    plt.imshow(image_)
    plt.title(f'pillow image {image_.size}')

    plt.subplot(1,2,2)
    plt_image = mpimg.imread(image_path)
    plt.imshow(plt_image)
    plt.title(f'matplotlib image {plt_image.shape}')

    print('Image size: ',image_.size)
    print('Image mode: ',image_.mode)
    print('Height',image_.height)
    print('width',image_.width)
    #image as numpy array
    image_array = np.asarray(image_)
    print('Shape of Image as per numpy', image_array.shape)
	
	
	

def pairplot(df, figsize = (20,20), hue:str = None):
    num_cols = df.select_dtypes('number').columns
    print(num_cols)

    total_cols = len(num_cols)
    plot_num = 1

    plt.figure(figsize=(20,20))
    for col in num_cols:
        for col2 in num_cols:
            plt.subplot(total_cols, total_cols, plot_num)
            plot_num += 1

            if col == col2:
                sns.histplot(x = df[col], hue = df[hue])

            else:
                sns.scatterplot(x = df[col], y = df[col2], hue = df[hue])
				
			
			

def make_confusion_matrix(y_true, y_pred, classes=None, figsize=(10, 10), text_size=15, norm=False, savefig=False):

    """Makes a labelled confusion matrix comparing predictions and ground truth labels.
  If classes is passed, confusion matrix will be labelled, if not, integer class values
  will be used.
  Args:
    y_true: Array of truth labels (must be same shape as y_pred).
    y_pred: Array of predicted labels (must be same shape as y_true).
    classes: Array of class labels (e.g. string form). If `None`, integer labels are used.
    figsize: Size of output figure (default=(10, 10)).
    text_size: Size of output figure text (default=15).
    norm: normalize values or not (default=False).
    savefig: save confusion matrix to file (default=False).
  
  Returns:
    A labelled confusion matrix plot comparing y_true and y_pred.
  Example usage:
    make_confusion_matrix(y_true=test_labels, # ground truth test labels
                          y_pred=y_preds, # predicted labels
                          classes=class_names, # array of class label names
                          figsize=(15, 15),
                          text_size=10)
  """  
    # Create the confustion matrix
    cm = confusion_matrix(y_true, y_pred)
    cm_norm = cm.astype("float") / cm.sum(axis=1)[:, np.newaxis] # normalize it
    n_classes = cm.shape[0] # find the number of classes we're dealing with

    # Plot the figure and make it pretty
    fig, ax = plt.subplots(figsize=figsize)
    cax = ax.matshow(cm, cmap=plt.cm.Blues) # colors will represent how 'correct' a class is, darker == better
    fig.colorbar(cax)

      # Are there a list of classes?
    labels = classes or np.arange(cm.shape[0])
    # Label the axes
    ax.set(title="Confusion Matrix",
           xlabel="Predicted label",
           ylabel="True label",
           xticks=np.arange(n_classes), # create enough axis slots for each class
           yticks=np.arange(n_classes), 
           xticklabels=labels, # axes will labeled with class names (if they exist) or ints
           yticklabels=labels)

    # Make x-axis labels appear on bottom
    ax.xaxis.set_label_position("bottom")
    ax.xaxis.tick_bottom()

    # Set the threshold for different colors
    threshold = (cm.max() + cm.min()) / 2.

    # Plot the text on each cell
    for i, j in itertools.product(range(cm.shape[0]), range(cm.shape[1])):
      if norm:
        plt.text(j, i, f"{cm[i, j]} ({cm_norm[i, j]*100:.1f}%)",
                horizontalalignment="center",
                color="white" if cm[i, j] > threshold else "black",
                size=text_size)
      else:
        plt.text(j, i, f"{cm[i, j]}",
                horizontalalignment="center",
                color="white" if cm[i, j] > threshold else "black",
                size=text_size)

    # Save the figure to the current working directory
    if savefig:
      fig.savefig("confusion_matrix.png")



def plot_history(history, metrics=['loss', 'accuracy'], colors=None, markers=None, linestyles=None, figsize=(15, 4)):
    """
    Plot training history metrics using Matplotlib.

    Args:
        history (History): History object returned by model.fit.
        metrics (list): List of metrics to plot (e.g., ['loss', 'accuracy']).
        colors (list): List of colors for corresponding metrics.
        markers (list): List of markers for corresponding metrics.
        linestyles (list): List of line styles for corresponding metrics.
        figsize (tuple): Figure size (width, height).
    """
    if colors is None:
        colors = ['blue', 'red', 'orange', 'green', 'purple', 'brown', 'pink', 'gray']
    if markers is None:
        markers = ['o', 's', 'D', '^', 'v', '<', '>', 'p']
    if linestyles is None:
        linestyles = ['-', '--', '-.', ':']


    for metric_idx, metric in enumerate(metrics):
        plt.figure(figsize=figsize)
        plt.title(f'{metric.capitalize()}')
        plt.grid(True)

        for i, metric_type in enumerate(['train', 'val']):
            metric_name = metric if metric_type == 'train' else f'val_{metric}'

            data = history.history.get(metric_name, None)
            if data is not None:
                plt.plot(range(1, len(data) + 1), data,
                         label=f'{metric_type.capitalize()} {metric}',
                         color=colors[i % len(colors)],
                         marker=markers[i % len(markers)],
                         linestyle=linestyles[i % len(linestyles)])

        plt.xlabel('Epoch')
        plt.ylabel(metric.capitalize())
        plt.legend()
        
    plt.show()




def compare_histories(histories:list, plot = ['loss','accuracy'], split = ['train','val'], epoch:int = None, figsize = (20,10), colors = None, **plot_kwargs ):

    ''' Compares Histories
	
	Arguments:
	###############
	histroies	:	list of Histories to compare
	plot:list	:   what to plot (what metrics you want to compare)  -> ['loss', 'accuracy']  
	split:list  :   what split to compare -> ['train', 'val']
	epoch:int   :   for how many epochs to comapre (cannot be greater than highest epoch of histories)
	colors		: 	list of colors for each histories to represent
	figsize:tuple:  tuple of size
	plot_kwargs :   kwargs to plt.plot to customize plot
	
	Returns:
	##############
	Plots plot of len(plot) * len(split) of comparing histories 
	
	'''
    if colors is not None and len(colors) != len(histories):
        print('Number of Histories and number of Colors are not equal')
        colors = None

    try:
        cols = []
        for i in plot:
            for j in split:
                if j == 'val':
                    cols.append(f'{j}_{i}')
                else:
                    cols.append(i)

        #compare to epoch
        if epoch is None:
            max_epoch = sorted(histories, key = (lambda x:max(x.epoch)), reverse=True)[0]
            epoch = max(max_epoch.epoch)+1

        def display(col, plot_num, history, epoch:int = None, **plot_kwargs):
            plt.subplot(len(plot),len(split),plot_num)
            plt.grid(True)

            if epoch is None:
                epoch = history.epoch

            plt.plot(np.arange(epoch), pd.DataFrame(history.history)[col], label=history.model.name, **plot_kwargs)
            plt.title((' '.join(col.split('_'))).upper())
            plt.xlabel('epochs')
            plt.ylabel(col.split('_')[-1])
            plt.legend()

        plt.figure(figsize = figsize)
        plot_title = " ".join(plot).upper()+" PLOT"
        plt.suptitle(plot_title)

        for plot_num,col in enumerate(cols,1):

            if colors is None:
                for hist in histories:
                    display(col, plot_num, hist, epoch)

            else:
                for hist,color in zip(histories,colors):
                    display(col, plot_num, hist, epoch, color = color)

    except Exception as e:
        print('Error Occured: ',e)



def distplot_axis(x:np.array, axis:list=list(range(95,100)), percent:int = 99,where_text:int = 5000,xlim:list=None, ylim:list=None, **kwargs):
    # plt.figure(figsize = (15,7))
    plt.grid()

    plt.hist(x , bins = 50, **kwargs)
    # Title Word Count distribution
    try:
        plt.xlim(xlim[0], xlim[1])
        plt.ylim(ylim[0], ylim[1])
    except:
        pass

    plt.xlabel('word length')
    plt.ylabel('Count')
    for i in axis:
        plt.axvline(x = np.percentile(x, i), color = 'b', label = 'axvline - full height')
        # random_pixel = 
        try:
            plt.text(np.percentile(x, i),where_text,f'{i} percent',rotation=90)
        except:
            pass

    plt.title(f'Text Word Count distribution: {np.percentile(x, percent)} words cover {percent}% of text data')






def plot_grid(*plots, grid_shape=(2, 2), figsize=(10, 8), **kwargs):
    num_plots = len(plots)
    rows, cols = grid_shape

    # Adjust figure size based on grid shape
    figsize = (figsize[0] * cols, figsize[1] * rows)

    fig, axes = plt.subplots(rows, cols, figsize=figsize, **kwargs)

    # Iterate over the plots and assign them to the corresponding axis
    for i, plot in enumerate(plots):
        ax = axes[i // cols, i % cols]  # Get the corresponding axis
        ax.plot(plot)  # Plot the data

    # Remove empty subplots if necessary
    if num_plots < rows * cols:
        for i in range(num_plots, rows * cols):
            fig.delaxes(axes[i // cols, i % cols])

    plt.tight_layout()  # Adjust spacing between subplots
    plt.show()







def main():
	pass

if __name__=="__main__":
	
	main()
